"""lmdierct"""
import asyncio
import logging

from lmdirect.const import DISABLED, ENABLED, MACHINE_NAME, MODEL_NAME, SERIAL_NUMBER

from .connection import Connection
from .msgs import (
    AUTO_BITFIELD,
    AUTO_BITFIELD_MAP,
    AUTO_SCHED_MAP,
    DOSE_K1,
    DOSE_TEA,
    FIRMWARE_VER,
    GLOBAL_AUTO,
    MODEL_GS3_AV,
    MODEL_LM,
    MON_OFF,
    MON_ON,
    MSGS,
    PREBREWING_TOFF_K1,
    PREBREWING_TON_K1,
    TSET_COFFEE,
    TSET_STEAM,
    TYPE_AUTO_ON_OFF,
    TYPE_COFFEE_TEMP,
    TYPE_MAIN,
    TYPE_PREBREW,
    TYPE_STEAM_TEMP,
    Msg,
)

_LOGGER = logging.getLogger(__name__)


class LMDirect(Connection):
    def __init__(self, machine_info):
        super().__init__(machine_info)

    @property
    def current_status(self):
        """Return a dict of all the properties that have been received"""
        return self._current_status

    @property
    def machine_name(self):
        """Return the name of the machine"""
        return self._machine_info[MACHINE_NAME]

    @property
    def serial_number(self):
        """Return serial number"""
        return self._machine_info[SERIAL_NUMBER]

    @property
    def model_name(self):
        """Return model name"""
        return self._machine_info[MODEL_NAME]

    @property
    def firmware_version(self):
        """Return firmware version"""
        return self._current_status.get(FIRMWARE_VER, "Unknown")

    def register_callback(self, callback):
        """Register callback for updates"""
        if callable(callback):
            self._callback_list.append(callback)

    def register_raw_callback(self, msg, callback, **kwargs):
        """Register a callback for the raw response to a command"""
        if callable(callback):
            self._raw_callback_list.append((msg, callback))

    def deregister_raw_callback(self, key):
        """Register a callback for the raw response to a command"""

        """The key is a tuple comprised of (msg, callback)"""
        if key in self._raw_callback_list:
            self._raw_callback_list.remove(key)

    async def connect(self):
        return await self._connect()

    async def request_status(self):
        """Request all status elements"""
        msgs = [
            Msg.GET_STATUS,
            Msg.GET_CONFIG,
            Msg.GET_AUTO_SCHED,
            Msg.GET_DATETIME,
            Msg.GET_DRINK_STATS,
        ]

        _LOGGER.debug("Requesting status")
        await asyncio.gather(*[self._send_msg(msg) for msg in msgs])

        """Also wait for current temp"""
        self._responses_waiting.append(MSGS[Msg.GET_TEMP_REPORT].msg)

    async def send_msg(self, msg_id, key=None, data=None):
        """Send a message"""
        await self._send_msg(msg_id, key=key, data=data)

    async def close(self):
        """Wait for the read task to exit"""
        if self._read_reaper_task:
            await asyncio.gather(self._read_reaper_task)
        await self._close()

    def call_callbacks(self, **kwargs):
        """Call all callbacks to refresh data"""
        self._call_callbacks(**kwargs)

    """Utils"""

    def convert_to_ascii(self, value, size):
        """Convert an integer value to ASCII-encoded hex"""
        return ("%0" + str(size * 2) + "X") % value

    def findkey(self, find_value, dict):
        """Find a key from the value in a dict"""
        return next(
            (key for key, value in dict.items() if value == find_value),
            None,
        )

    """Services"""

    async def set_power(self, power):
        """Send power on or power off commands"""
        value = self.convert_to_ascii(0x01 if power else 0x00, size=1)
        await self._send_msg(Msg.SET_POWER, data=value)
        await self._send_msg(Msg.GET_CONFIG)
        self._call_callbacks(entity_type=TYPE_MAIN)

    async def set_auto_on_off(self, day_of_week=None, enable=None):
        """Configure auto on/off"""

        async def auto_on_off_callback(key, data):
            """Callback for the raw sequence"""

            """We don't need to continue looking for it"""
            self.deregister_raw_callback(key)
            _LOGGER.debug(f"Received raw callback: {data}")

            try:
                """Find the bit"""
                elem = self.findkey(AUTO_BITFIELD, AUTO_SCHED_MAP)

                """The strings are ASCII-encoded hex, so each value takes 2 bytes"""
                index = elem.index * 2
                size = elem.size * 2

                """Extract value for this field"""
                orig_value = int(data[index : index + size], 16)
                bitmask = 0x01 << self.findkey(day_of_week, AUTO_BITFIELD_MAP)
                new_value = self.convert_to_ascii(
                    orig_value | bitmask if enable else orig_value & ~bitmask, 1
                )

                buf_to_send = data[:index] + new_value + data[index + size :]

                self._current_status[day_of_week] = ENABLED if enable else DISABLED

                await self._send_msg(Msg.SET_AUTO_SCHED, data=buf_to_send)
                await self._send_msg(Msg.GET_AUTO_SCHED)

                self._call_callbacks(entity_type=TYPE_AUTO_ON_OFF)
            except Exception as err:
                _LOGGER.debug(f"Caught exception: {err}")

        if None in [day_of_week, enable]:
            msg = f"Some parameters invalid {day_of_week} {enable}"
            raise InvalidInput(msg)

        self.register_raw_callback(Msg.GET_AUTO_SCHED, auto_on_off_callback)
        await self._send_msg(Msg.GET_AUTO_SCHED)

    async def set_auto_on_off_global(self, value):
        """Set global auto on/off"""
        await self.set_auto_on_off(GLOBAL_AUTO, value)

    async def set_auto_on_off_hours(
        self, day_of_week=None, hour_on=None, hour_off=None
    ):
        """Configure auto on/off"""

        async def auto_on_off_callback(key, data):
            """Callback for the raw sequence"""

            """We don't need to continue looking for it"""
            self.deregister_raw_callback(key)
            _LOGGER.debug(f"Received raw callback: {data}")

            try:
                """Find the "on" element"""
                elem = self.findkey(day_of_week + MON_ON[3:], AUTO_SCHED_MAP)

                """The strings are ASCII-encoded hex, so each value takes 2 bytes and there 2 of them"""
                index = elem.index * 2
                size = elem.size * 4

                """Construct the new "on" and "off" values"""
                buf_to_send = (
                    data[:index]
                    + self.convert_to_ascii(hour_on, 1)
                    + self.convert_to_ascii(hour_off, 1)
                    + data[index + size :]
                )

                self._current_status[day_of_week + MON_ON[3:]] = hour_on
                self._current_status[day_of_week + MON_OFF[3:]] = hour_off

                await self._send_msg(Msg.SET_AUTO_SCHED, data=buf_to_send)
                await self._send_msg(Msg.GET_AUTO_SCHED)
                self._call_callbacks(entity_type=TYPE_AUTO_ON_OFF)
            except Exception as err:
                _LOGGER.debug(f"Caught exception: {err}")
                raise

        if None in [day_of_week, hour_on, hour_off]:
            msg = f"Some parameters invalid {day_of_week} {hour_on} {hour_off}"
            raise InvalidInput(msg)

        self.register_raw_callback(Msg.GET_AUTO_SCHED, auto_on_off_callback)
        await self._send_msg(Msg.GET_AUTO_SCHED)

    async def set_dose(self, key=None, pulses=None):
        """Set the coffee boiler temp in Celcius"""

        if None in [key, pulses]:
            msg = f"set_dose: Some parameters not specified {key} {pulses}"
            raise InvalidInput(msg)

        if isinstance(pulses, str):
            pulses = int(pulses)

        if isinstance(key, str):
            key = int(key)

        """Validate input"""
        if not (1 <= pulses <= 1000 and 1 <= key <= 5):
            msg = f"set_dose: Invalid values pulses:{pulses} key:{key}"
            raise InvalidInput(msg)

        self._current_status[DOSE_K1.replace("1", str(key))] = pulses

        data = self.convert_to_ascii(pulses, size=2)
        key = self.convert_to_ascii(Msg.DOSE_KEY_BASE + (key - 1) * 2, size=1)
        await self._send_msg(Msg.SET_DOSE, key=key, data=data)
        await self._send_msg(Msg.GET_CONFIG)
        self._call_callbacks(entity_type=TYPE_MAIN)

    async def set_dose_tea(self, seconds=None):
        """Set the coffee boiler temp in Celcius"""

        if not seconds:
            raise InvalidInput("set_dose_tea: Seconds not specified")

        if isinstance(seconds, str):
            seconds = int(seconds)

        """Validate input"""
        if not (1 <= seconds <= 30):
            msg = f"Invalid values seconds:{seconds}"
            raise InvalidInput(msg)

        self._current_status[DOSE_TEA] = seconds

        data = self.convert_to_ascii(seconds, size=1)
        await self._send_msg(Msg.SET_DOSE_TEA, data=data)
        await self._send_msg(Msg.GET_CONFIG)
        self._call_callbacks(entity_type=TYPE_MAIN)

    async def set_prebrew_times(self, key=None, time_on=None, time_off=None):
        """Set the coffee boiler temp in Celcius"""

        if None in [key, time_on, time_off]:
            raise InvalidInput(
                "set_prebrew_times: Some parameters invalid {key} {time_on} {time_off}"
            )

        if isinstance(time_on, str):
            time_on = float(time_on)

        if isinstance(time_off, str):
            time_off = float(time_off)

        if isinstance(key, str):
            key = int(key)

        """Validate input"""
        if not (0 <= time_on <= 5.9 and 0 <= time_off <= 5.9 and 1 <= key <= 4):
            msg = f"Invalid values time_on:{time_on} off_time:{time_off}"
            raise InvalidInput(msg)

        if not (
            (self.model_name == MODEL_GS3_AV and 1 <= key <= 4)
            or (self.model_name == MODEL_LM and key == 1)
        ):
            msg = f"Invalid values key:{key}"
            raise InvalidInput(msg)

        self._current_status[PREBREWING_TON_K1.replace("1", str(key))] = time_on
        self._current_status[PREBREWING_TOFF_K1.replace("1", str(key))] = time_off

        """set "on" time"""
        key_on = self.convert_to_ascii(Msg.PREBREW_ON_BASE + (key - 1), size=1)
        data = self.convert_to_ascii(int(time_on * 10), size=1)
        await self._send_msg(Msg.SET_PREBREW_TIMES, key=key_on, data=data)

        """set "off" time"""
        key_off = self.convert_to_ascii(Msg.PREBREW_OFF_BASE + (key - 1), size=1)
        data = self.convert_to_ascii(int(time_off * 10), size=1)
        await self._send_msg(Msg.SET_PREBREW_TIMES, key=key_off, data=data)
        await self._send_msg(Msg.GET_CONFIG)
        self._call_callbacks(entity_type=TYPE_PREBREW)

    async def set_coffee_temp(self, temp=None):
        """Set the coffee boiler temp in Celcius"""

        if not temp:
            raise InvalidInput("set_coffee__temp: Temperature not specified")

        if isinstance(temp, str):
            temp = float(temp)

        temp = round(temp, 1)

        self._current_status[TSET_COFFEE] = temp

        data = self.convert_to_ascii(int(temp * 10), size=2)
        await self._send_msg(Msg.SET_COFFEE_TEMP, data=data)
        await self._send_msg(Msg.GET_TEMP_REPORT, data=data)
        self._call_callbacks(entity_type=TYPE_COFFEE_TEMP)

    async def set_steam_temp(self, temp=None):
        """Set the steam boiler temp in Celcius"""

        if not temp:
            raise InvalidInput("set_steam_temp: Temperature not specified")

        if isinstance(temp, str):
            temp = float(temp)

        temp = round(temp, 1)

        self._current_status[TSET_STEAM] = temp

        data = self.convert_to_ascii(int(temp * 10), size=2)
        await self._send_msg(Msg.SET_STEAM_TEMP, data=data)
        await self._send_msg(Msg.GET_TEMP_REPORT, data=data)
        self._call_callbacks(entity_type=TYPE_STEAM_TEMP)

    async def set_prebrewing_enable(self, enable):
        """Turn prebrewing on or off"""
        data = self.convert_to_ascii(0x01 if enable else 0x00, size=1)
        await self._send_msg(Msg.SET_PREBREWING_ENABLE, data=data)
        await self._send_msg(Msg.GET_CONFIG)
        self._call_callbacks(entity_type=TYPE_PREBREW)


class InvalidInput(BaseException):
    """Error to indicate there is no Connection"""

    def __init__(self, msg):
        _LOGGER.error(msg)


import base64
import logging
from math import ceil

from Crypto.Cipher import AES

_LOGGER = logging.getLogger(__name__)


class AESCipher:
    def __init__(self, key):
        self.key = key.encode("latin-1")

    def encrypt(self, plaintext):
        cipher = AES.new(self.key, AES.MODE_CBC, iv=bytearray(16))
        plaintext = bytes(plaintext, "utf-8") + bytearray(
            ceil(len(plaintext) / 16) * 16 - len(plaintext)
        )
        ciphertext = cipher.encrypt(plaintext)
        return base64.b64encode(ciphertext)

    def decrypt(self, b64text):
        ciphertext = base64.b64decode(b64text)
        cipher = AES.new(self.key, AES.MODE_CBC, iv=bytearray(16))
        return cipher.decrypt(ciphertext).decode().partition("\0")[0]



"""Constants for the lmdirect package."""

ENABLED = "Enabled"
DISABLED = "Disabled"

"""API endpoints for the La Marzocco CMS and gateway."""
GW_URL = "https://gw.lamarzocco.io/v1/home/machines"
TOKEN_URL = "https://cms.lamarzocco.io/oauth/v2/token"
CUSTOMER_URL = "https://cms.lamarzocco.io/api/customer"
DRINK_COUNTER_URL = (
    "https://gw.lamarzocco.io/v1/home/machines/{serial_number}/statistics/counters"
)
UPDATE_URL = (
    "https://gw.lamarzocco.io/v1/home/machines/updates-available?device=machine"
)

"""Config parameters."""
HOST = "host"
PORT = "port"
CLIENT_ID = "client_id"
CLIENT_SECRET = "client_secret"
USERNAME = "username"
PASSWORD = "password"
KEY = "key"
SERIAL_NUMBER = "serial_number"
MACHINE_NAME = "machine_name"
MODEL_NAME = "model_name"
DRINKS_K1_OFFSET = "drinks_k1_offset"
DRINKS_K2_OFFSET = "drinks_k2_offset"
DRINKS_K3_OFFSET = "drinks_k3_offset"
DRINKS_K4_OFFSET = "drinks_k4_offset"
CONTINUOUS_OFFSET = "continuous_offset"
HOT_WATER_OFFSET = "hot_water_offset"

CONFIG_PARAMS = [
    HOST,
    PORT,
    CLIENT_ID,
    CLIENT_SECRET,
    USERNAME,
    PASSWORD,
    KEY,
    SERIAL_NUMBER,
    MACHINE_NAME,
    MODEL_NAME,
    DRINKS_K1_OFFSET,
    DRINKS_K2_OFFSET,
    DRINKS_K3_OFFSET,
    DRINKS_K4_OFFSET,
    CONTINUOUS_OFFSET,
    HOT_WATER_OFFSET,
]



"""lmdirect connection class"""
import asyncio
import logging
from datetime import datetime, timedelta
from functools import partial

from authlib.integrations.base_client.errors import OAuthError
from authlib.integrations.httpx_client import AsyncOAuth2Client

from .aescipher import AESCipher
from .const import *
from .msgs import (
    AUTO_BITFIELD,
    AUTO_BITFIELD_MAP,
    DIVIDE_KEYS,
    DRINK_OFFSET_MAP,
    FIRMWARE_VER,
    GATEWAY_DRINK_MAP,
    MSGS,
    SERIAL_NUMBERS,
    UPDATE_AVAILABLE,
    Elem,
    Msg,
)

_LOGGER = logging.getLogger(__name__)


class Connection:
    def __init__(self, machine_info):
        """Init LMDirect"""
        self._reader = None
        self._writer = None
        self._read_response_task = None
        self._read_reaper_task = None
        self._current_status = {}
        self._responses_waiting = []
        self._run = True
        self._callback_list = []
        self._raw_callback_list = []
        self._cipher = None
        self._machine_info = machine_info
        self._start_time = None
        self._connected = False
        self._lock = asyncio.Lock()
        self._first_time = True
        self._update_available = None

    async def retrieve_machine_info(self, machine_info):
        """Machine data inialization"""
        _LOGGER.debug(f"Retrieving machine info")

        client = AsyncOAuth2Client(
            client_id=machine_info[CLIENT_ID],
            client_secret=machine_info[CLIENT_SECRET],
            token_endpoint=TOKEN_URL,
        )

        headers = {
            "client_id": machine_info[CLIENT_ID],
            "client_secret": machine_info[CLIENT_SECRET],
        }

        try:
            await client.fetch_token(
                url=TOKEN_URL,
                username=machine_info[USERNAME],
                password=machine_info[PASSWORD],
                headers=headers,
            )
        except OAuthError:
            await client.aclose()
            raise AuthFail("Authorization failure")

        except Exception as err:
            await self.close()
            raise AuthFail(f"Caught: {type(err)}, {err}")

        """Only retrieve this the first time"""
        if any(
            x not in machine_info
            for x in [KEY, SERIAL_NUMBER, MACHINE_NAME, MODEL_NAME]
        ):
            cust_info = await client.get(CUSTOMER_URL)
            if cust_info:
                fleet = cust_info.json()["data"]["fleet"][0]
                machine_info[KEY] = fleet["communicationKey"]
                machine_info[SERIAL_NUMBER] = fleet["machine"]["serialNumber"]
                machine_info[MACHINE_NAME] = fleet["name"]
                machine_info[MODEL_NAME] = fleet["machine"]["model"]["name"]

        if any(x not in self._current_status for x in DRINK_OFFSET_MAP.values()):
            drink_info = await client.get(
                DRINK_COUNTER_URL.format(serial_number=machine_info[SERIAL_NUMBER])
            )
            if drink_info:
                data = drink_info.json()["data"]
                self._current_status.update(
                    {GATEWAY_DRINK_MAP[x["coffeeType"]]: x["count"] for x in data}
                )

        if UPDATE_AVAILABLE not in self._current_status:
            update_info = await client.get(UPDATE_URL)
            if update_info:
                data = update_info.json()["data"]
                self._update_available = "Yes" if data else "No"

        """Done with the cloud API"""
        await client.aclose()

        _LOGGER.debug(f"Finished machine info")
        return machine_info

    async def _connect(self):
        """Conmnect to espresso machine"""
        if self._connected:
            return self._machine_info

        _LOGGER.debug(f"Connecting")

        try:
            self._machine_info = await self.retrieve_machine_info(self._machine_info)
        except Exception as err:
            raise ConnectionFail(f"Exception retrieving machine info: {err}")

        if not self._cipher:
            self._cipher = AESCipher(self._machine_info[KEY])

        """Connect to the machine"""
        try:
            self._reader, self._writer = await asyncio.wait_for(
                asyncio.open_connection(
                    self._machine_info[HOST], self._machine_info[PORT]
                ),
                timeout=3,
            )
        except asyncio.TimeoutError:
            _LOGGER.warning("Connection Timeout, skipping")
            return None

        except Exception as err:
            raise ConnectionFail(f"Cannot connect to machine: {err}")

        """Start listening for responses & sending status requests"""
        await self.start_read_task()

        self._connected = True
        _LOGGER.debug("Finished Connecting")

        return self._machine_info

    async def start_read_task(self):
        """Start listening for responses & sending status requests"""
        loop = asyncio.get_event_loop()
        self._read_response_task = loop.create_task(
            self.read_response_task(), name="Response Task"
        )

        """Reap the results and any any exceptions"""
        self._read_reaper_task = loop.create_task(
            self.read_reaper(), name="Read Reaper"
        )

    async def _close(self):
        """Close the connection"""
        if not self._connected:
            return

        if self._writer is not None:
            self._writer.close()

        if self._read_response_task:
            self._read_response_task.cancel()

        self._reader = self._writer = None
        self._connected = False
        _LOGGER.debug("Finished closing")

    async def read_reaper(self):
        _LOGGER.debug("Starting read reaper")
        try:
            await asyncio.gather(self._read_response_task)
        except Exception as err:
            _LOGGER.error(f"Exception in read_response_task: {err}")

        await self._close()
        self._read_response_task = None
        self._first_time = False

        _LOGGER.debug("Finished reaping read task")

    def _call_callbacks(self, **kwargs):
        """Call the callbacks"""
        if self._callback_list is not None:
            [
                elem(
                    current_status=self._current_status,
                    **kwargs,
                )
                for elem in self._callback_list
            ]

    async def read_response_task(self):
        """Start thread to receive responses"""
        BUFFER_SIZE = 1000
        handle = None

        _LOGGER.debug("Starting read task")

        if self._start_time is None:
            self._start_time = datetime.now()

        while self._run:
            encoded_data = await self._reader.read(BUFFER_SIZE)
            if encoded_data is not None:
                loop = asyncio.get_event_loop()
                fn = partial(self._cipher.decrypt, encoded_data[1:-1])
                plaintext = await loop.run_in_executor(None, fn)
                if not plaintext:
                    continue

                await self.process_data(plaintext)

                if not self._first_time:
                    if handle:
                        handle.cancel()
                        handle = None

                    """Throttle callbacks"""
                    handle = loop.call_later(5, self._call_callbacks)
                else:
                    self._call_callbacks()

                """Exit if we've been reading longer than 5s"""
                if datetime.now() > self._start_time + timedelta(seconds=5):
                    _LOGGER.debug(f"Exiting loop: {self._responses_waiting}")

                    """Flush the wait list"""
                    self._responses_waiting = []
                    break

    async def process_data(self, plaintext):
        """Process incoming packet"""

        """Separate the mesg from the data"""
        msg_type = plaintext[0]
        raw_data = plaintext[1:]
        msg = raw_data[:8]

        """chop off the message and check byte"""
        data = raw_data[8:-2]
        finished = not len(self._responses_waiting)

        _LOGGER.debug(f"Message={msg}, Data={data}")

        if msg_type == Msg.WRITE:
            if Msg.RESPONSE_GOOD not in data:
                _LOGGER.error(f"Command Failed: {msg}: {data}")
            else:
                _LOGGER.debug(f"Command Succeeded: {msg}: {data}")
        else:
            """Find the matching item or returns None"""
            msg_id = next(
                (
                    x
                    for x in MSGS
                    if MSGS[x].msg == msg and MSGS[x].msg_type == msg_type
                ),
                None,
            )

            if msg_id is None:
                _LOGGER.error(f"Unexpected response: {plaintext}")
                return False
            else:
                cur_msg = MSGS[msg_id]

            """notify any listeners for this message"""
            [
                await x[1]((msg_id, x[1]), data)
                for x in self._raw_callback_list
                if MSGS[x[0]].msg == msg
            ]

            if cur_msg.map is not None:
                await self._populate_items(data, cur_msg)

        if msg in self._responses_waiting:
            self._responses_waiting.remove(msg)
            finished = not len(self._responses_waiting)
            if finished:
                _LOGGER.debug("Received all responses")

        return True

    async def _populate_items(self, data, cur_msg):
        """process all the fields"""
        map = cur_msg.map
        for elem in map:
            """The strings are ASCII-encoded hex, so each value takes 2 bytes"""
            index = elem.index * 2
            size = elem.size * 2

            """Extract value for this field"""
            value = data[index : index + size]

            if elem.type == Elem.INT:
                """Convert from ascii-encoded hex"""
                value = int(value, 16)

            if any(x in map[elem] for x in DIVIDE_KEYS):
                value = value / 10
            elif map[elem] == FIRMWARE_VER:
                value = "%0.2f" % (value / 100)
            elif map[elem] in SERIAL_NUMBERS:
                value = "".join(
                    [chr(int(value[i : i + 2], 16)) for i in range(0, len(value), 2)]
                )
                value = value.partition("\0")[0]

            elif map[elem] == AUTO_BITFIELD:
                for item in AUTO_BITFIELD_MAP:
                    setting = ENABLED if value & 0x01 else DISABLED
                    self._current_status[AUTO_BITFIELD_MAP[item]] = setting
                    value = value >> 1
                continue
            elif map[elem] in DRINK_OFFSET_MAP:
                offset_index = DRINK_OFFSET_MAP[map[elem]]
                if map[elem] not in self._current_status:
                    """If we haven't seen the value before, calculate the offset"""
                    self._current_status.update(
                        {offset_index: value - self._current_status[offset_index]}
                    )
                """Offset the value"""
                value -= self._current_status[offset_index]

            self._current_status[map[elem]] = value

    async def _send_msg(self, msg_id, data=None, key=None):
        """Send command to espresso machine"""

        def checksum(buffer):
            """Compute check byte"""
            buffer = bytes(buffer, "utf-8")
            return "%0.2X" % (sum(buffer) % 256)

        """Prevent race conditions - can be called from different tasks"""
        async with self._lock:
            msg = MSGS[msg_id]
            _LOGGER.debug(f"Sending {msg.msg} with {data}")

            """Connect if we don't have an active connection"""
            result = await self._connect()

            if not result:
                raise ConnectionFail("Connection failed")

            if not self._writer:
                raise ConnectionFail(f"self._writer={self._writer}")

            """If a key was provided, replace the second byte of the message"""
            msg_to_send = msg.msg if not key else msg.msg[:2] + key + msg.msg[4:]

            if key:
                msg_to_send = msg_to_send[:2] + key + msg_to_send[4:]

            plaintext = msg.msg_type + msg_to_send

            if data is not None:
                plaintext += data

            """Add the check byte"""
            plaintext += checksum(plaintext)

            loop = asyncio.get_event_loop()
            fn = partial(self._cipher.encrypt, plaintext)
            ciphertext = (
                "@" + (await loop.run_in_executor(None, fn)).decode("utf-8") + "%"
            )

            self._writer.write(bytes(ciphertext, "utf-8"))
            await self._writer.drain()

            """Remember that we're waiting for a response"""
            self._responses_waiting.append(msg_to_send)

            """Note when the command was sent"""
            self._start_time = datetime.now()

            _LOGGER.debug("Finished sending")


class AuthFail(BaseException):
    """Error to indicate there is invalid auth"""

    def __init__(self, msg):
        _LOGGER.error(msg)


class ConnectionFail(BaseException):
    """Error to indicate there is no Connection"""

    def __init__(self, msg):
        _LOGGER.error(msg)



"""Local API commands for La Marzocco espresso machines."""

"""All models currently Wifi-enabled."""
MODEL_GS3_AV = "GS3 AV"
MODEL_GS3_MP = "GS3 MP"
MODEL_LM = "Linea Mini"

# Tags

from lmdirect.const import (
    CONTINUOUS_OFFSET,
    DRINKS_K1_OFFSET,
    DRINKS_K2_OFFSET,
    DRINKS_K3_OFFSET,
    DRINKS_K4_OFFSET,
    HOT_WATER_OFFSET,
)

DATE_RECEIVED = "date_received"
POWER = "power"
TEMP_COFFEE = "coffee_temp"
TEMP_STEAM = "steam_temp"
TSET_STEAM = "steam_set_temp"
TSET_COFFEE = "coffee_set_temp"
DOSE_K1 = "dose_k1"
DOSE_K2 = "dose_k2"
DOSE_K3 = "dose_k3"
DOSE_K4 = "dose_k4"
DOSE_K5 = "dose_k5"
DOSE_TEA = "dose_tea"
ENABLE_PREBREWING = "enable_prebrewing"
PREBREWING_TON_K1 = "prebrewing_ton_k1"
PREBREWING_TON_K2 = "prebrewing_ton_k2"
PREBREWING_TON_K3 = "prebrewing_ton_k3"
PREBREWING_TON_K4 = "prebrewing_ton_k4"
PREBREWING_TOFF_K1 = "prebrewing_toff_k1"
PREBREWING_TOFF_K2 = "prebrewing_toff_k2"
PREBREWING_TOFF_K3 = "prebrewing_toff_k3"
PREBREWING_TOFF_K4 = "prebrewing_toff_k4"
DRINKS_K1 = "drinks_k1"
DRINKS_K2 = "drinks_k2"
DRINKS_K3 = "drinks_k3"
DRINKS_K4 = "drinks_k4"
CONTINUOUS = "continuous"
TOTAL_COFFEE = "total_coffee"
HOT_WATER = "hot_water"
DRINK_MYSTERY = "drink_mystery"
TOTAL_DRINKS = "total_drinks"
HOT_WATER_2 = "hot_water_2"
DRINKS_TEA = "drinks_tea"

FIRMWARE_VER = "firmware_ver"
MODULE_SER_NUM = "module_ser_num"

UPDATE_AVAILABLE = "update_available"

SUN_ON = "sun_on"
SUN_OFF = "sun_off"
MON_ON = "mon_on"
MON_OFF = "mon_off"
TUE_ON = "tue_on"
TUE_OFF = "tue_off"
WED_ON = "wed_on"
WED_OFF = "wed_off"
THU_ON = "thu_on"
THU_OFF = "thu_off"
FRI_ON = "fri_on"
FRI_OFF = "fri_off"
SAT_ON = "sat_on"
SAT_OFF = "sat_off"

AUTO_BITFIELD = "auto_bitfield"
GLOBAL_AUTO = "global_auto"
SUN_AUTO = "sun"
MON_AUTO = "mon"
TUE_AUTO = "tue"
WED_AUTO = "wed"
THU_AUTO = "thu"
FRI_AUTO = "fri"
SAT_AUTO = "sat"

SECOND = "second"
MINUTE = "minute"
HOUR = "hour"
DAYOFWEEK = "dayofweek"
DAY = "day"
MONTH = "month"
YEAR = "month"

FLOW_MYSTERY = "flow_mystery"
FLOW_PULSES = "flow_pulses"
FLOW_SECONDS = "flow_seconds"

WRITE_RESULT = "write_result"
MACHINE_SER_NUM = "machine_ser_num"

"""Data Types"""
TYPE_MAIN = 1
TYPE_PREBREW = 2
TYPE_AUTO_ON_OFF = 3
TYPE_COFFEE_TEMP = 4
TYPE_STEAM_TEMP = 5
TYPE_DRINK_STATS = 6

DIVIDE_KEYS = ["temp", "prebrewing_to"]
SERIAL_NUMBERS = [MACHINE_SER_NUM, MODULE_SER_NUM]

# Response Maps
class Elem:

    INT = 0
    STRING = 1

    def __init__(self, index, size=1, type=INT):
        self._index = index
        self._size = size
        self._type = type

    @property
    def index(self):
        return self._index

    @property
    def size(self):
        return self._size

    @property
    def type(self):
        return self._type


# R
# 40 1C 00 04: Message
# 03 BC: Coffee Temp (95.6C)
# 04 D8: Steam Temp (124.0C)
# B6: Check byte

TEMP_REPORT_MAP = {Elem(0, 2): TEMP_COFFEE, Elem(2, 2): TEMP_STEAM}

# R
# 40 00 00 20: Message
# 01
# 78: Version (1.20)
# 02
# 53 6E XX XX XX XX XX XX XX XX XX XX: Module serial number (SnXXXXXXXXXX)
# 00: ?? sometimes 01
# 00 00 00 00 00 00 00
# 27: 01: Power (ON)
# 01 00 00 00
# 32: 02 96: Coffee Temp (66.2C)
# 34: 03 2A: Steam Temp (81.0C)
# 70: Check byte

STATUS_MAP = {
    Elem(1, 1): FIRMWARE_VER,
    Elem(3, 12, type=Elem.STRING): MODULE_SER_NUM,
    Elem(23, 1): POWER,
    # Elem(32, 2): TEMP_COFFEE,
    # Elem(34, 2): TEMP_STEAM,
}

# R
# 00 00 00 1F: Preamble
# 01 00 00 02 6E 31 39
# 11: 03 C7: Coffee Temp Set (96.7C)
# 13: 04 D9: Steam Temp Set (124.1C)
# 15: 00: Prebrewing ON (off)
# 16: 00: B1 Prebrewing ON
# 17: 00: B2 Prebrewing ON
# 18: 00: B3 Prebrewing ON
# 19: 0A: B4 Prebrewing ON (1.0s)
# 20: 00: B1 Prebrewing Off
# 21: 00: B2 Prebrewing Off
# 22: 00: B3 Prebrewing Off
# 23: 00: B4 Prebrewing Off
# 24: 0078: Dose B1
# 26: 0076: Dost B2
# 28: 0078: Dose B3
# 30: 0085: Dose B4
# 32: 03E8: Dose B5
# 34: 08: Seconds hot water for Tea
# 35: Check byte

CONFIG_MAP = {
    Elem(7, 2): TSET_COFFEE,
    Elem(9, 2): TSET_STEAM,
    Elem(11): ENABLE_PREBREWING,
    Elem(12): PREBREWING_TON_K1,
    Elem(13): PREBREWING_TON_K2,
    Elem(14): PREBREWING_TON_K3,
    Elem(15): PREBREWING_TON_K4,
    Elem(16): PREBREWING_TOFF_K1,
    Elem(17): PREBREWING_TOFF_K2,
    Elem(18): PREBREWING_TOFF_K3,
    Elem(19): PREBREWING_TOFF_K4,
    Elem(20, 2): DOSE_K1,
    Elem(22, 2): DOSE_K2,
    Elem(24, 2): DOSE_K3,
    Elem(26, 2): DOSE_K4,
    Elem(28, 2): DOSE_K5,
    Elem(30): DOSE_TEA,
}

# Response to R 03 10 00 1D EB

# R
# 03 10 00 1D: Message
# 4: FE: Bitfield for auto on/off, where upper bits are the days bit 0 is auto on/off (disabled)
# 5: 06: Sunday on hour
# 6: 11: Sunday off hour (5pm)
# 7: 06: Monday on hour
# 8: 11: Monday off hour (5pm)
# 9: 06: Tuesday on hour
# 10: 11: Tuesday off hour (5pm)
# 11: 06: Wednesday on hour
# 12: 11: Wednesday off hour (5pm)
# 13: 06: Thursday on hour
# 14: 11: Thursday off hour (5pm)
# 15: 06: Friday on hour
# 16: 11: Friday off hour (5pm)
# 17: 06: Saturday on hour
# 18: 11: Saturday off hour (5pm)
# 00 00 00 00 00 00 00 00 00 00 00 00 00 00
# 2E: Check byte

AUTO_SCHED_MAP = {
    Elem(0): AUTO_BITFIELD,
    Elem(1): SUN_ON,
    Elem(2): SUN_OFF,
    Elem(3): MON_ON,
    Elem(4): MON_OFF,
    Elem(5): TUE_ON,
    Elem(6): TUE_OFF,
    Elem(7): WED_ON,
    Elem(8): WED_OFF,
    Elem(9): THU_ON,
    Elem(10): THU_OFF,
    Elem(11): FRI_ON,
    Elem(12): FRI_OFF,
    Elem(13): SAT_ON,
    Elem(14): SAT_OFF,
}

AUTO_BITFIELD_MAP = {
    0: GLOBAL_AUTO,
    1: MON_AUTO,
    2: TUE_AUTO,
    3: WED_AUTO,
    4: THU_AUTO,
    5: FRI_AUTO,
    6: SAT_AUTO,
    7: SUN_AUTO,
}

# W: Write
# 00 00 00 01 : Message
# OK: Result
# 72: Check byte

# WRITE_RESULT_MAP = {
#     Elem(0, 1, type=Elem.STRING): WRITE_RESULT,
# }

# App: R 01 00 00 11 D5
# Machine:
# R
# 01 00 00 11: Read machine serial number
# 5A: Z (?)
# XX XX XX XX XX XX XX XX 00 00 00 00 00 00 00 00: Machine serial number
# 7D: Check byte

SER_NUM_MAP = {
    Elem(1, 16, type=Elem.STRING): MACHINE_SER_NUM,
}

# R
# 03000007
# 00: Second (00)
# 05: Minute (5)
# 0A: Hour (10)
# 02: Day of week (2)
# 16: Day (22)
# 0C: Month (12)
# 14: Year (20)
# B3: Check byte

DATETIME_MAP = {
    Elem(0, 1): SECOND,
    Elem(1, 1): MINUTE,
    Elem(2, 1): HOUR,
    Elem(3, 1): DAYOFWEEK,
    Elem(4, 1): DAY,
    Elem(5, 1): MONTH,
    Elem(6, 1): YEAR,
}

# R
# 0020002C
# 0000014A: Key 1 (1 espresso)
# 00000096: Key 2 (2 espressos) * 2 = 0x12C
# 000001B0: Key 3 (1 coffee)
# 00000021: Key 4 (2 coffees) * 2 = 0x42
# 00000563: Flushing
# 00000914: From Coffee boiler (total of above)
# 00000010: Tea
# 0000000A: ?? Doesn't seem to change
# 00000ADE: ?? Increments every button press
# 00000010: Also Tea?
# 00000032: # Tea, if you let it finish

DRINK_STATS_MAP = {
    Elem(0, 4): DRINKS_K1,
    Elem(4, 4): DRINKS_K2,
    Elem(8, 4): DRINKS_K3,
    Elem(12, 4): DRINKS_K4,
    Elem(16, 4): CONTINUOUS,
    Elem(20, 4): TOTAL_COFFEE,
    Elem(24, 4): HOT_WATER,
    Elem(28, 4): DRINK_MYSTERY,
    Elem(32, 4): TOTAL_DRINKS,
    Elem(36, 4): HOT_WATER_2,
    Elem(40, 4): DRINKS_TEA,
}

GATEWAY_DRINK_MAP = {
    -1: HOT_WATER_OFFSET,
    0: DRINKS_K1_OFFSET,
    1: DRINKS_K2_OFFSET,
    2: DRINKS_K3_OFFSET,
    3: DRINKS_K4_OFFSET,
    4: CONTINUOUS_OFFSET,
}

DRINK_OFFSET_MAP = {
    HOT_WATER: HOT_WATER_OFFSET,
    DRINKS_K1: DRINKS_K1_OFFSET,
    DRINKS_K2: DRINKS_K2_OFFSET,
    DRINKS_K3: DRINKS_K3_OFFSET,
    DRINKS_K4: DRINKS_K4_OFFSET,
    CONTINUOUS: CONTINUOUS_OFFSET,
}

# Z
# 60000016
# Snxxxxxxxxxx
# 0401: Mystery
# 0009: Pulses
# 000C: Seconds (upper 12 bits are whole seconds, lower 4 bits are fractional)
# 03B6: Coffee boiler temp
# 04D4: Steam boiler temp
# DC: Check byte

WATER_FLOW_MAP = {
    Elem(12, 2): FLOW_MYSTERY,
    Elem(14, 2): FLOW_PULSES,
    Elem(16, 2): FLOW_SECONDS,
    # Elem(18, 2): TEMP_COFFEE,
    # Elem(20, 2): TEMP_STEAM,
}


class Msg:
    GET_STATUS = 0
    GET_CONFIG = 1
    GET_AUTO_SCHED = 2
    SET_POWER = 3
    GET_TEMP_REPORT = 4
    SET_AUTO_ON = 4
    SET_AUTO_OFF = 5
    GET_SER_NUM = 6
    GET_DATETIME = 7
    SET_COFFEE_TEMP = 8
    SET_STEAM_TEMP = 9
    SET_PREBREWING_ENABLE = 10
    SET_AUTO_SCHED = 11
    SET_DOSE = 12
    SET_DOSE_TEA = 13
    SET_PREBREW_TIMES = 14
    GET_DRINK_STATS = 15
    GET_WATER_FLOW = 16

    """Dose command starts at 0x14 for the first key and increments by 2 thereafter"""
    DOSE_KEY_BASE = 0x14

    """Prebrew on base"""
    PREBREW_ON_BASE = 0x0C

    """Prebrew off base"""
    PREBREW_OFF_BASE = 0x10

    READ = "R"
    WRITE = "W"
    STREAM = "Z"

    RESPONSE_GOOD = "OK"

    def __init__(self, msg_type, msg, map):
        """Init command"""
        self._msg_type = msg_type
        self._msg = msg
        self._map = map

    @property
    def msg_type(self):
        """Command type"""
        return self._msg_type

    @property
    def msg(self):
        """Command type"""
        return self._msg

    @property
    def map(self):
        """Return map if we should decode, None if not"""
        return self._map


MSGS = {
    # Reads
    Msg.GET_STATUS: Msg(Msg.READ, "40000020", STATUS_MAP),
    Msg.GET_CONFIG: Msg(Msg.READ, "0000001F", CONFIG_MAP),
    Msg.GET_AUTO_SCHED: Msg(Msg.READ, "0310001D", AUTO_SCHED_MAP),
    Msg.GET_TEMP_REPORT: Msg(Msg.READ, "401C0004", TEMP_REPORT_MAP),
    Msg.GET_SER_NUM: Msg(Msg.READ, "01000011", SER_NUM_MAP),
    Msg.GET_DATETIME: Msg(Msg.READ, "03000007", DATETIME_MAP),
    Msg.GET_DRINK_STATS: Msg(Msg.READ, "0020002C", DRINK_STATS_MAP),
    Msg.GET_WATER_FLOW: Msg(Msg.STREAM, "60000016", WATER_FLOW_MAP),
    # Writes
    Msg.SET_POWER: Msg(Msg.WRITE, "00000001", None),
    Msg.SET_COFFEE_TEMP: Msg(Msg.WRITE, "00070002", None),
    Msg.SET_STEAM_TEMP: Msg(Msg.WRITE, "00090002", None),
    Msg.SET_PREBREWING_ENABLE: Msg(Msg.WRITE, "000B0001", None),
    Msg.SET_AUTO_SCHED: Msg(Msg.WRITE, "0310001D", None),
    # Write config for keys (second byte is base + the key number and will be replaced)
    Msg.SET_DOSE: Msg(Msg.WRITE, "00140002", None),
    Msg.SET_DOSE_TEA: Msg(Msg.WRITE, "001E0001", None),
    Msg.SET_PREBREW_TIMES: Msg(Msg.WRITE, "000C0001", None),
}



import setuptools

with open("README.md", "r") as f:
    readme = f.read()

setuptools.setup(
    name="lmdirect",
    version="0.6.2",
    description="A Python implementation of the local La Marzocco API",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/rccoleman/lmdirect",
    author="Rob Coleman",
    author_email="rccoleman@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    packages=setuptools.find_packages(),
    install_requires=["pycryptodome>=3.9.9", "httpx>=0.16.1", "authlib>=0.15.2"],
    package_data={
        "license": ["LICENSE"],
    },
)
